
module Main where

import ATermLib hiding (parseOptions)
import Language.VDM.VDMLib
import System.Console.GetOpt
import System.Exit (exitWith, exitFailure, ExitCode(..))
import System.Environment (getArgs, getProgName)
import Control.Monad (when, mapM_)
import IO (hPutStrLn, hFlush, stderr)
import SGLR (sglr)


-- * Data structures for parsing options


-- ** Data Types

-- | Type that holds all options required by the program.
data Opt = Opt
   { -- | VDM source filename that will be used for input
     optInput      :: String
     -- | Function that will output the final result
   , optOutput     :: String -> IO ()
     -- | Conversion function that will be applied before outputting results
   , optConversion :: Document -> String
   }


-- ** Initial Values

-- | Initial program options that will be used if no options are passed to
-- the program.
--
--    [@optInput@]     The input file will be initialized with the empty string,
--                     although this is not required because it will be
--                     validated (but removes the warning from the ghc compiler)
--
--    [@optOutput@]     Function that will output the result. If no option is
--                      given the output will be printed in the stdout.
--
--    [@optConversion@] Function that will convert the AST in a string to be
--                      printed. If no option is given, the default function
--                      pretty-prints the AST.
--
startOpt :: Opt
startOpt = Opt
   { optInput      = ""
   , optOutput     = putStr
   , optConversion = renderVDM
   }


-- | Description of all available options
--
--    [@h@] Show help with usage information
--
--    [@i@] Input file to be read for parsing
--
--    [@o@] Output file to all results will be saved
--
--    [@a | p@] Alternative options that allow to control the output format.
--              If the @a@ option is supplied then the Document's AST is
--              printed, otherwise, if the @p@ is supplied, the Documet is
--              pretty-printed
--
options :: [OptDescr (Opt -> IO Opt)]
options =
   [ Option "h" ["help"] 
        (NoArg (\opt -> exitHelp)) 
        "Show usage info"
   , Option "i" ["input"] 
        (ReqArg 
           (\arg opt -> do
              return opt { optInput = arg })
           "FILE")        
        "Input file (mandatory)"
   , Option "o" ["output"]
        (ReqArg
           (\arg opt -> do
              return opt { optOutput = writeFile arg })
           "FILE")
        "Output file (default: stdout)"
   , Option "a" ["ast"]
        (NoArg (\opt -> do return opt { optConversion = showDocumentAST }))
        "Print Abstract Syntax Tree"
   , Option "p" ["pretty-printer"]
        (NoArg (\opt -> do return opt { optConversion = renderVDM }))
        "Print using the pretty printer (default)"
   ]


-- * Main Function

-- | Main function that provides interaction with the system.
main :: IO ()
main = do 
   (opts, _) <- parseOptions
   let Opt { optInput      = input
           , optOutput     = output 
           , optConversion = convert 
           } = opts
   doc <- sglr "../syn/isovdm.def.tbl" input "Document"
   (output . convert) doc


-- * Auxiliar Functions

-- | Function that prints the Abstract Syntax Tree of a 'Document'. This use the
-- show function provided by the generated instances from DrIFT.
showDocumentAST :: Document -> String
showDocumentAST = show


-- | Function that reads arguments form the environment and parses it according
-- with rules specified in the 'options' type.
parseOptions :: IO (Opt, [String])
parseOptions = do
   -- Reads the options to a triple
   (optsActions, rest, errs) <- getArgs >>= return . getOpt RequireOrder options

   -- If there are errors show them, show the usage help and exit with failure
   -- code
   when (not (null errs)) $ do
      mapM_ (hPutStrLn stderr) errs
      showHelp
      exitFailure
   
   -- If no options were specified then show the help and exit with failure code
   when (null optsActions) $ do
      showHelp
      exitFailure
            
   opts <- foldl (>>=) (return startOpt) optsActions
   
   -- verify if the -i option was specified. This must be done, since this
   -- option is mandatory
   when (optInput opts == "") $ do
      exitErrorHelp "An input file is required"
      
   -- return all the parsed options   
   return (opts, rest)


-- | Function that prints the help usage and returns with a success code (used
-- when the help program option is specified.
exitHelp :: IO a
exitHelp = do
    showHelp
    exitWith ExitSuccess


-- | Function that prints the program usage to the sdtderr using the standard
-- 'usageInfo' function.
showHelp :: IO ()
showHelp = do
    prg <- getProgName
    hPutStrLn stderr (usageInfo prg options)
    hFlush stderr


-- | Helper function that prints a message to stderr and exits with a failure
-- code error
exitErrorHelp :: String -> IO ()
exitErrorHelp msg = do
   hPutStrLn stderr ("Error: " ++ msg)
   hPutStrLn stderr ""
   showHelp
   exitFailure

